package flow

//
// Find a way to improve this mess, maybe it can be merged in one func
//
//

import (
	"errors"
	"fmt"
	"reflect"
	"sync"
)

// Operation interface
type Operation interface { // Id perhaps?
	ID() string
	Set(input Data) // Special var method
	Process(params ...Data) (Data, error)
}

// Will be named operation
type operation struct {
	sync.Mutex
	flow     *Flow
	id       string
	name     string
	kind     string
	inputs   []*operation // still figuring, might be Operation
	setter   func(Data)
	executor interface{}
}

// OpCtx operation Context
type OpCtx = *sync.Map

// NewOpCtx creates a running context
func newOpCtx() OpCtx {
	return &sync.Map{}
}

func (o *operation) ID() string { return o.id }

func (o *operation) Process(params ...Data) (Data, error) {
	// Create CTX
	ctx := newOpCtx()
	return executeOP(o.flow, o.id, ctx, params...)
}
func (o *operation) Set(data Data) {
	if o.setter != nil {
		o.setter(data)
	}
}

// make Executor for func

// The function that executions an operation
func executeOP(f *Flow, id string, ctx OpCtx, params ...Data) (ret Data, err error) {
	defer func() {
		if r := recover(); r != nil {
			err = fmt.Errorf("%v", r)
			f.hooks.error(id, err)
		}
	}()

	op, ok := f.getOp(id)
	if !ok {
		err = fmt.Errorf("invalid operation '%s'", id)
		f.hooks.error(id, err)
		return nil, err
	}
	// Simple data returner
	switch fn := op.executor.(type) {
	case func() Data:
		f.hooks.start(id)
		res := fn()
		f.hooks.finish(id, res)
		return res, nil
	case func(...Data) Data:
		f.hooks.start(id)
		res := fn(params...)
		f.hooks.finish(id, res)
		return res, nil

	}
	// Input param

	/*if f.err != nil {
		return nil
	}*/
	op.Lock()
	defer op.Unlock()
	if ctx != nil {
		if v, ok := ctx.Load(id); ok { // Cache
			return v, nil
		}
	}

	// Check inputs
	fnval := reflect.ValueOf(op.executor)
	if fnval.Type().NumIn() != len(op.inputs) {
		err = fmt.Errorf("expect %d inputs got %d", fnval.Type().NumIn(), len(op.inputs))
		f.hooks.error(id, err)
		return nil, err
	}
	/////////////////////////////
	// NEW PARALLEL PROCESSING
	///////////
	f.hooks.wait(id)

	callParam := make([]reflect.Value, len(op.inputs))

	// Parallel processing if inputs
	wg := sync.WaitGroup{}
	wg.Add(len(op.inputs))
	for i, in := range op.inputs {
		go func(i int, in *operation) {
			defer wg.Done()
			fr, err := executeOP(f, in.id, ctx, params...)
			if err != nil {
				callParam[i] = reflect.Value{}
				return
			}

			if fr == nil {
				callParam[i] = reflect.Zero(fnval.Type().In(i))
				return
			}

			callParam[i] = reflect.ValueOf(fr)
		}(i, in)
	}
	wg.Wait()
	// Return type error checking
	//
	errMsg := ""
	for i, p := range callParam {
		if !p.IsValid() {
			//callParam[i] = reflect.Zero(fnval.Type().In(i))
			errMsg += fmt.Sprintf("Input %d invalid\n", i)
		} else if !p.Type().AssignableTo(fnval.Type().In(i)) {
			errMsg += fmt.Sprintf("Input %d type mismatch expected: %v got :%v\n", i, fnval.Type().In(i), p.Type())
		}
	}
	if len(errMsg) > 0 {
		err := errors.New(errMsg)
		f.hooks.error(id, err)
		return nil, err
	}

	// The actual operation process
	f.hooks.start(id)

	fnret := fnval.Call(callParam)
	if len(fnret) > 1 && (fnret[len(fnret)-1].Interface() != nil) {
		err, ok := fnret[1].Interface().(error)
		if !ok {
			err = errors.New("unknown error")
		}
		f.hooks.error(id, err)
		return nil, err
	}

	// THE RESULT
	ret = fnret[0].Interface()
	if ctx != nil {
		ctx.Store(id, ret)
	}
	f.hooks.finish(id, ret)
	return ret, nil
}
